"""
.. module:: theories.camb

:Synopsis: Managing the CAMB cosmological code
:Author: Jesus Torrado and Antony Lewis

.. |br| raw:: html

   <br />

This module imports and manages the CAMB cosmological code.
It requires CAMB 1.0 or higher (for compatibility with older versions, you can temporarily
use cobaya 1.0.4, but update asap, since that version is not maintained any more).

.. note::

   **If you use this cosmological code, please cite it as:**
   |br|
   A. Lewis, A. Challinor, A. Lasenby,
   *Efficient computation of CMB anisotropies in closed FRW*
   (`arXiv:astro-ph/9911177 <https://arxiv.org/abs/astro-ph/9911177>`_)
   |br|
   C. Howlett, A. Lewis, A. Hall, A. Challinor,
   *CMB power spectrum parameter degeneracies in the era of precision cosmology*
   (`arXiv:1201.3654 <https://arxiv.org/abs/1201.3654>`_)


Usage
-----

If you are using a likelihood that requires some observable from CAMB, simply add CAMB
to the theory block.

You can specify any parameter that CAMB understands in the ``params`` block:

.. code-block:: yaml

   theory:
     camb:
       extra_args:
         [any param that CAMB understands, for FIXED and PRECISION]

   params:
       [any param that CAMB understands, fixed, sampled or derived]


If you want to use your own version of CAMB, you need to specify its location with a
``path`` option inside the ``classy`` block. If you do not specify a ``path``,
CAMB will be loaded from the automatic-install ``modules`` folder, if specified, or
otherwise imported as a globally-installed Python package. Cobaya will print at
initialisation where it is getting CAMB from.

.. _camb_modify:

Modifying CAMB
^^^^^^^^^^^^^^

If you modify CAMB and add new variables, make sure that the variables you create are
exposed in the Python interface (`instructions here
<https://camb.readthedocs.io/en/latest/model.html#camb.model.CAMBparams>`__).
If you follow those instructions you do not need to make any additional modification in
Cobaya.

You can use the :doc:`model wrapper <cosmo_model>` to test your modification by
evaluating observables or getting derived quantities at known points in the parameter
space (set ``debug: True`` to get more detailed information of what exactly is passed to
CLASS).

In your CAMB modification, remember that you can raise a ``CAMBParamRangeError`` or a
``CAMBError`` whenever the computation of any observable would fail, but you do not
expect that observable to be compatible with the data (e.g. at the fringes of the
parameter space). Whenever such an error is raised during sampling, the likelihood is
assumed to be zero, and the run is not interrupted.


Installation
------------

Pre-requisites
^^^^^^^^^^^^^^

**cobaya** calls CAMB using its Python interface, which requires that you compile CAMB
using intel's ifort compiler or the GNU gfortran compiler version 6.4 or later.
To check if you have the latter, type ``gfortran --version`` in the shell,
and the first line should look like

.. code::

   GNU Fortran ([your OS version]) [gfortran version] [release date]

Check that ``[gfortran's version]`` is at least 6.4. If you get an error instead, you need
to install gfortran (contact your local IT service).

CAMB comes with binaries pre-built for Windows, so if you don't need to modify the CAMB
source code, no Fortran compiler is needed.


Automatic installation
^^^^^^^^^^^^^^^^^^^^^^

If you do not plan to modify CAMB, the easiest way to install it is using the
:doc:`automatic installation script <installation_cosmo>`. Just make sure that
``theory: camb:`` appears in one of the files passed as arguments to the installation
script.


Manual installation (or using your own version)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you are planning to modify CAMB or use an already modified version,
you should not use the automatic installation script. Use the installation method that
best adapts to your needs:

* [**Recommended for staying up-to-date**]
  To install CAMB locally and keep it up-to-date, clone the
  `CAMB repository in Github <https://github.com/cmbant/CAMB>`_
  in some folder of your choice, say ``/path/to/theories/CAMB``:

  .. code:: bash

      $ cd /path/to/theories
      $ git clone --recursive https://github.com/cmbant/CAMB.git
      $ cd CAMB
      $ python setup.py build

  To update to the last changes in CAMB (master), run ``git pull`` from CAMB's folder and
  re-build using the last command.

* [**Recommended for modifying CAMB**]
  First, `fork the CAMB repository in Github <https://github.com/cmbant/CAMB>`_
  (follow `this instructions <https://help.github.com/articles/fork-a-repo/>`_) and then
  follow the same steps as above, substituting the second one with:

  .. code:: bash

      $ git clone --recursive https://[YourGithubUser]@github.com/[YourGithubUser]/CAMB.git

* To use your own version, assuming it's placed under ``/path/to/theories/CAMB``,
  just make sure it is compiled (and that the version on top of which you based your
  modifications is old enough to have the Python interface implemented.

In the cases above, you **must** specify the path to your CAMB installation in
the input block for CAMB (otherwise a system-wide CAMB may be used instead):

.. code:: yaml

   theory:
     camb:
       path: /path/to/theories/CAMB

.. note::

   In any of these methods, you should **not** install CAMB as python package using
   ``python setup.py install --user``, as the official instructions suggest.
   It is actually safer not to do so if you intend to switch between different versions or
   modifications of CAMB.
"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import sys
import os
import logging
from copy import deepcopy
import numpy as np
from collections import namedtuple, OrderedDict as odict
from time import time
from numbers import Number

# Local
from cobaya.theories._cosmo import _cosmo
from cobaya.log import HandledException
from cobaya.install import download_github_release, check_gcc_version
from cobaya.conventions import _c_km_s, _T_CMB_K

# Result collector
collector = namedtuple("collector", ["method", "args", "kwargs"])
collector.__new__.__defaults__ = (None, [], {})


class camb(_cosmo):

    def initialize(self):
        """Importing CAMB from the correct path, if given."""
        if not self.path and self.path_install:
            self.path = os.path.join(
                self.path_install, "code", camb_repo_name[camb_repo_name.find("/") + 1:])
        if self.path and not os.path.exists(self.path):
            # Fail if this was a directly specified path,
            # or ignore and try to global-import if it came from a path_install
            if self.path_install:
                self.log.info("*local* CAMB not found at " + self.path)
                self.log.info("Importing *global* CAMB.")
            else:
                self.log.error("*local* CAMB not found at " + self.path)
                raise HandledException
        elif self.path:
            self.log.info("Importing *local* CAMB from " + self.path)
            if not os.path.exists(self.path):
                self.log.error("The given folder does not exist: '%s'", self.path)
                raise HandledException
            pycamb_path = self.path
            if not os.path.exists(os.path.join(self.path, "setup.py")):
                self.log.error(
                    "Either CAMB is not in the given folder, '%s', or you are using a "
                    "very old version without the Python interface.", self.path)
                raise HandledException
            sys.path.insert(0, pycamb_path)
        else:
            self.log.info("Importing *global* CAMB.")
        try:
            import camb
        except ImportError:
            self.log.error(
                "Couldn't find the CAMB python interface.\n"
                "Make sure that you have compiled it, and that you either\n"
                " (a) specify a path (you didn't) or\n"
                " (b) install the Python interface globally with\n"
                "     'pip install -e /path/to/camb [--user]'")
            raise HandledException
        self.camb = camb
        # Prepare errors
        from camb.baseconfig import CAMBParamRangeError, CAMBError
        from camb.baseconfig import CAMBValueError, CAMBUnknownArgumentError
        global CAMBParamRangeError, CAMBError, CAMBValueError, CAMBUnknownArgumentError
        # Generate states, to avoid recomputing
        self.n_states = 3
        self.states = [{"params": None, "derived": None, "derived_extra": None, "last": 0}
                       for i in range(self.n_states)]
        # Dict of named tuples to collect requirements and computation methods
        self.collectors = {}
        # Additional input parameters to pass to CAMB, and attributes to set_ manually
        self.extra_args = self.extra_args or {}
        self.extra_attrs = {}
        # Patch: if cosmomc_theta is used, CAMB needs to be passed explicitly "H0=None"
        # This is *not* going to change on CAMB's side.
        if all((p in self.input_params) for p in ["H0", "cosmomc_theta"]):
            self.log.error("Can't pass both H0 and cosmomc_theta to Camb.")
            raise HandledException
        if "cosmomc_theta" in self.input_params:
            self.extra_args["H0"] = None
        # Set aliases
        self.planck_to_camb = self.renames
        # Derived parameters that may not have been requested, but will be necessary later
        self.derived_extra = []
        self.needs_perts = False

    def current_state(self):
        lasts = [self.states[i]["last"] for i in range(self.n_states)]
        return self.states[lasts.index(max(lasts))]

    def needs(self, **requirements):
        # Computed quantities required by the likelihood
        # Note that redshifts below are treated differently for background quantities,
        #   were no additional transfer computation is needed (e.g. H(z)),
        #   and matter-power-related quantities, that require additional computation
        #   and need the redshifts to be passed at CAMBParams instantiation.
        #   Also, we always make sure that those redshifts are sorted in descending order,
        #   since all CAMB related functions return quantities in that implicit order
        # The following super call makes sure that the requirements are properly
        # accumulated, i.e. taking the max of precision requests, etc.
        super(camb, self).needs(**requirements)
        for k, v in self._needs.items():
            # Products and other computations
            if k == "Cl":
                self.extra_args["lmax"] = max(
                    max(v.values()), self.extra_args.get("lmax", 0))
                cls = [a.lower() for a in v]
                self.collectors[k] = collector(
                    method="CAMBdata.get_cmb_power_spectra",
                    kwargs={
                        "spectra": list(set(
                            (self.collectors[k].kwargs.get("spectra", [])
                             if k in self.collectors else []) +
                            ["total"] + (["lens_potential"] if "pp" in cls else []))),
                        "raw_cl": True})
                self.needs_perts = True
            elif k == "H":
                self.collectors[k] = collector(
                    method="CAMBdata.h_of_z",
                    kwargs={"z": np.atleast_1d(v["z"])})
                self.H_units_conv_factor = {"1/Mpc": 1, "km/s/Mpc": _c_km_s}
            elif k == "angular_diameter_distance":
                self.collectors[k] = collector(
                    method="CAMBdata.angular_diameter_distance",
                    kwargs={"z": np.atleast_1d(v["z"])})
            elif k == "comoving_radial_distance":
                self.collectors[k] = collector(
                    method="CAMBdata.comoving_radial_distance",
                    kwargs={"z": np.atleast_1d(v["z"])})
            elif k == "fsigma8":
                self.add_to_redshifts(v["z"])
                self.collectors[k] = collector(
                    method="CAMBdata.get_fsigma8",
                    kwargs={})
                self.needs_perts = True
            elif k == "Pk_interpolator":
                self.extra_args["kmax"] = max(v["k_max"], self.extra_args.get("kmax", 0))
                self.add_to_redshifts(v["z"])
                v["vars_pairs"] = v["vars_pairs"] or [["total", "total"]]
                kwargs = deepcopy(v)
                # change of defaults:
                kwargs["hubble_units"] = kwargs.get("hubble_units", False)
                kwargs["k_hunit"] = kwargs.get("k_hunit", False)
                for p in "k_max", "z", "vars_pairs":
                    kwargs.pop(p)
                for pair in v["vars_pairs"]:
                    name = "Pk_interpolator_%s_%s" % (pair[0], pair[1])
                    kwargs.update(dict(zip(["var1", "var2"], pair)))
                    self.collectors[name] = collector(
                        method="CAMBdata.get_matter_power_interpolator",
                        kwargs=kwargs)
                self.needs_perts = True
            # General derived parameters
            elif v is None:
                k_translated = self.translate_param(k)
                if k_translated not in self.derived_extra:
                    self.derived_extra += [k_translated]
                if k == "sigma8":
                    self.extra_attrs["WantTransfer"] = True
                    self.needs_perts = True
            else:
                self.log.error("This should not be happening. Contact the developers.")
                raise HandledException
        # Finally, check that there are no repeated parameters between input and extra
        if set(self.input_params).intersection(set(self.extra_args)):
            self.log.error(
                "The following parameters appear both as input parameters and as CAMB "
                "extra arguments: %s. Please, remove one of the definitions of each.",
                list(set(self.input_params).intersection(set(self.extra_args))))
            raise HandledException
        # Remove extra args that cause an error if the associated product is not requested
        if "Cl" not in [k for k in self._needs]:
            self.extra_args.pop("lens_potential_accuracy", None)

    def add_to_redshifts(self, z):
        self.extra_args["redshifts"] = np.sort(np.unique(np.concatenate(
            (np.atleast_1d(z), self.extra_args.get("redshifts", [])))))[::-1]

    def translate_param(self, p):
        if self.use_planck_names:
            return self.planck_to_camb.get(p, p)
        return p

    def set(self, params_values_dict, i_state):
        # Store them, to use them later to identify the state
        self.states[i_state]["params"] = deepcopy(params_values_dict)
        # Prepare parameters to be passed: this-iteration + extra
        args = {self.translate_param(p): v for p, v in params_values_dict.items()}
        args.update(self.extra_args)
        self.states[i_state]["set_args"] = deepcopy(args)
        # Generate and save
        self.log.debug("Setting parameters: %r", args)
        try:
            cambparams = self.camb.set_params(**args)
            for attr, value in self.extra_attrs.items():
                if hasattr(cambparams, attr):
                    setattr(cambparams, attr, value)
                else:
                    self.log.error("Some of the attributes to be set manually were not "
                                   "recognized: %s=%s", attr, value)
                    raise HandledException
            return cambparams
        except CAMBParamRangeError:
            if self.stop_at_error:
                self.log.error("Out of bound parameters: %r", params_values_dict)
                raise HandledException
            else:
                self.log.debug("Out of bounds parameters. "
                               "Assigning 0 likelihood and going on.")
        except CAMBValueError:
            self.log.error(
                "Error setting parameters (see traceback below)! "
                "Parameters sent to CAMB: %r and %r.\n"
                "To ignore this kind of errors, make 'stop_at_error: False'.",
                self.states[i_state]["params"], self.extra_args)
            raise
        except CAMBUnknownArgumentError as e:
            self.log.error(
                "Some of the parameters passed to CAMB were not recognized: %s" % str(e))
            raise HandledException
        return False

    def compute(self, _derived=None, cached=True, **params_values_dict):
        lasts = [self.states[i]["last"] for i in range(self.n_states)]
        try:
            if not cached:
                raise StopIteration
            # are the parameter values there already?
            i_state = next(i for i in range(self.n_states)
                           if self.states[i]["params"] == params_values_dict)
            # has any new product been requested?
            for product in self.collectors:
                next(k for k in self.states[i_state] if k == product)
            reused_state = True
            # Get (pre-computed) derived parameters
            if _derived == {}:
                _derived.update(self.states[i_state]["derived"])
            self.log.debug("Re-using computed results (state %d)", i_state)
        except StopIteration:
            reused_state = False
            # update the (first) oldest one and compute
            i_state = lasts.index(min(lasts))
            self.log.debug("Computing (state %d)", i_state)
            if self.timing:
                a = time()
            # Set parameters
            result = self.set(params_values_dict, i_state)
            # Failed to set parameters but no error raised
            # (e.g. out of computationally feasible range): lik=0
            if not result:
                return 0
            intermediates = {
                "CAMBparams": {"result": result},
                "CAMBdata": {"method": "get_results" if self.needs_perts
                else "get_background",
                             "result": None, "derived_dic": None}}
            # Compute the necessary products (incl. any intermediate result, if needed)
            for product, collector in self.collectors.items():
                parent, method = collector.method.split(".")
                try:
                    if intermediates[parent]["result"] is None:
                        intermediates[parent]["result"] = getattr(
                            self.camb, intermediates[parent]["method"])(
                            intermediates["CAMBparams"]["result"])
                    method = getattr(intermediates[parent]["result"], method)
                    self.states[i_state][product] = method(
                        *self.collectors[product].args, **self.collectors[product].kwargs)
                except CAMBError:
                    if self.stop_at_error:
                        self.log.error(
                            "Computation error! Parameters sent to CAMB: %r and %r.\n"
                            "To ignore this kind of errors, make 'stop_at_error: False'.",
                            self.states[i_state]["params"], self.extra_args)
                        raise HandledException
                    else:
                        # Assumed to be a "parameter out of range" error.
                        return 0
            # Prepare derived parameters
            if _derived == {}:
                _derived.update(self._get_derived_all(intermediates))
                self.states[i_state]["derived"] = odict(
                    [[p, _derived[p]] for p in self.output_params])
            # Prepare necessary extra derived parameters
            self.states[i_state]["derived_extra"] = {
                p: self._get_derived(p, intermediates) for p in self.derived_extra}
            if self.timing:
                self.n += 1
                self.time_avg = (time() - a + self.time_avg * (self.n - 1)) / self.n
                self.log.debug("Average running time: %g seconds", self.time_avg)
        # make this one the current one by decreasing the antiquity of the rest
        for i in range(self.n_states):
            self.states[i]["last"] -= max(lasts)
        self.states[i_state]["last"] = 1
        return 1 if reused_state else 2

    def _get_derived_from_params(self, p, intermediates):
        for origin in ["CAMBdata", "CAMBparams"]:
            result = intermediates[origin].get("result")
            if result is None:
                continue
            for thing in [result, getattr(result, "Params", {})]:
                try:
                    return getattr(thing, p)
                except AttributeError:
                    for mod in ["InitPower", "Reion", "Recomb", "Transfer", "DarkEnergy"]:
                        try:
                            return getattr(getattr(thing, mod), p)
                        except AttributeError:
                            pass
        return None

    def _get_derived_from_std(self, p, intermediates):
        dic = intermediates["CAMBdata"].get("derived_dic", None)
        if dic is None:
            result = intermediates["CAMBdata"].get("result", None)
            if result is None:
                return None
            dic = result.get_derived_params()
            intermediates["CAMBdata"]["derived_dic"] = dic
        return dic.get(p, None)

    def _get_derived_from_getter(self, p, intermediates):
        return getattr(intermediates["CAMBparams"]["result"], "get_" + p, lambda: None)()

    def _get_derived(self, p, intermediates):
        """
        General function to extract a single derived parameter.

        To get a parameter *from a likelihood* use `get_param` instead.
        """
        # Specific calls, if general ones fail:
        if p == "sigma8":
            return intermediates["CAMBdata"]["result"].get_sigma8()[-1]
        for f in [self._get_derived_from_params,
                  self._get_derived_from_std,
                  self._get_derived_from_getter]:
            derived = f(p, intermediates)
            if derived is not None:
                return derived

    def _get_derived_all(self, intermediates):
        """
        Returns a dictionary of derived parameters with their values,
        using the *current* state (i.e. it should only be called from
        the ``compute`` method).

        To get a parameter *from a likelihood* use `get_param` instead.
        """
        derived = {}
        for p in self.output_params:
            derived[p] = self._get_derived(self.translate_param(p), intermediates)
            if derived[p] is None:
                self.log.error(
                    "Derived param '%s' not implemented in the CAMB interface", p)
                raise HandledException
        return derived

    def get_param(self, p):
        current_state = self.current_state()
        for pool in ["params", "derived", "derived_extra"]:
            value = deepcopy(
                (current_state[pool] or {}).get(self.translate_param(p), None))
            if value is not None:
                return value
        self.log.error("Parameter not known: '%s'", p)
        raise HandledException

    def get_cl(self, ell_factor=False, units="muK2"):
        current_state = self.current_state()
        # get C_l^XX from the cosmological code
        try:
            cl_camb = deepcopy(current_state["Cl"]["total"])
        except:
            self.log.error(
                "No Cl's were computed. Are you sure that you have requested them?")
            raise HandledException
        mapping = {"tt": 0, "ee": 1, "bb": 2, "te": 3}
        cls = {"ell": np.arange(cl_camb.shape[0])}
        cls.update({sp: cl_camb[:, i] for sp, i in mapping.items()})
        if "lens_potential" in current_state["Cl"]:
            cls.update({"pp": deepcopy(current_state["Cl"]["lens_potential"])[:, 0]})
        # unit conversion and ell_factor
        ell_factor = ((cls["ell"] + 1) * cls["ell"] / (2 * np.pi))[2:] if ell_factor else 1
        units_factors = {"1": 1,
                         "muK2": _T_CMB_K * 1.e6,
                         "K2": _T_CMB_K}
        try:
            units_factor = units_factors[units]
        except KeyError:
            self.log.error("Units '%s' not recognized. Use one of %s.",
                           units, list(units_factors))
            raise HandledException
        for cl in cls:
            if cl not in ['pp', 'ell']:
                cls[cl][2:] *= units_factor ** 2 * ell_factor
        if 'pp' in cls and ell_factor is not 1:
            cls['pp'][2:] *= ell_factor ** 2 * (2 * np.pi)
        return cls

    def _get_z_dependent(self, quantity, z):
        if quantity == "fsigma8":
            computed_redshifts = self.extra_args["redshifts"]
        else:
            z_name = next(k for k in ["redshifts", "z"]
                          if k in self.collectors[quantity].kwargs)
            computed_redshifts = self.collectors[quantity].kwargs[z_name]
        i_kwarg_z = np.concatenate(
            [np.where(computed_redshifts == zi)[0] for zi in np.atleast_1d(z)])
        return np.array(deepcopy(self.current_state()[quantity]))[i_kwarg_z]

    def get_H(self, z, units="km/s/Mpc"):
        try:
            return self._get_z_dependent("H", z) * self.H_units_conv_factor[units]
        except KeyError:
            self.log.error("Units not known for H: '%s'. Try instead one of %r.",
                           units, list(self.H_units_conv_factor))
            raise HandledException

    def get_angular_diameter_distance(self, z):
        return self._get_z_dependent("angular_diameter_distance", z)

    def get_comoving_radial_distance(self, z):
        return self._get_z_dependent("comoving_radial_distance", z)

    def get_fsigma8(self, z):
        return self._get_z_dependent("fsigma8", z)

    def get_Pk_interpolator(self):
        current_state = self.current_state()
        prefix = "Pk_interpolator_"
        return {k[len(prefix):]: deepcopy(v)
                for k, v in current_state.items() if k.startswith(prefix)}


# Installation routines ##################################################################

# Name of the Class repo/folder and version to download
camb_repo_name = "cmbant/CAMB"
camb_repo_version = "1.0.7"
camb_min_gcc_version = "6.4"


def is_installed(**kwargs):
    import platform
    if not kwargs["code"]:
        return True
    return os.path.isfile(os.path.realpath(
        os.path.join(
            kwargs["path"], "code", camb_repo_name[camb_repo_name.find("/") + 1:],
            "camb", "cambdll.dll" if (platform.system() == "Windows") else "camblib.so")))


def install(path=None, force=False, code=True, no_progress_bars=False, **kwargs):
    log = logging.getLogger(__name__.split(".")[-1])
    if not code:
        log.info("Code not requested. Nothing to do.")
        return True
    gcc_check = check_gcc_version(camb_min_gcc_version, error_returns=-1)
    if gcc_check == -1:
        log.warn("Failed to get gcc version (maybe not using gcc?). "
                 "Going ahead and hoping for the best.")
    elif not gcc_check:
        log.error("CAMB requires a gcc version >= %s, "
                  "which is higher than your current one.", camb_min_gcc_version)
        raise HandledException
    log.info("Downloading camb...")
    success = download_github_release(
        os.path.join(path, "code"), camb_repo_name, camb_repo_version,
        no_progress_bars=no_progress_bars)
    if not success:
        log.error("Could not download camb.")
        return False
    camb_path = os.path.join(path, "code", camb_repo_name[camb_repo_name.find("/") + 1:])
    log.info("Compiling camb...")
    from subprocess import Popen, PIPE
    process_make = Popen([sys.executable, "setup.py", "build_cluster"],
                         cwd=camb_path, stdout=PIPE, stderr=PIPE)
    out, err = process_make.communicate()
    if process_make.returncode:
        log.info(out)
        log.info(err)
        log.error("Compilation failed!")
        return False
    return True



"""
.. module:: likelihood

:Synopsis: Prototype likelihood and likelihood manager
:Author: Jesus Torrado

This module defines the main :class:`Likelihood` class, from which every likelihood
inherits, and the :class:`LikelihoodCollection` class, which groups and manages all the
individual likelihoods and is the actual instance passed to the sampler.

"""
# Python 2/3 compatibility
from __future__ import absolute_import, division

# Global
import sys
import traceback
from collections import OrderedDict as odict
from time import time, sleep
import numpy as np
from itertools import chain, permutations
import six
from copy import deepcopy

if six.PY3:
    from math import gcd
else:
    from fractions import gcd

# Local
from cobaya.conventions import _external, _theory, _params, _overhead_per_param
from cobaya.conventions import _timing, _p_renames, _chi2, _separator
from cobaya.conventions import _input_params, _output_params
from cobaya.conventions import _input_params_prefix, _output_params_prefix
from cobaya.tools import get_class, get_external_function, getargspec
from cobaya.tools import compare_params_lists
from cobaya.log import HandledException

# Logger
import logging

# Default options for all subclasses
class_options = {"speed": -1}


class Likelihood(object):
    """Likelihood class prototype."""

    # Generic initialization -- do not touch
    def __init__(self, info, modules=None, timing=None):
        self.name = self.__class__.__name__
        self.log = logging.getLogger(self.name)
        self.path_install = modules
        # Load info of the likelihood
        for k in info:
            setattr(self, k, info[k])
        # States, to avoid recomputing
        self.n_states = 3
        self.states = [{"params": None, "logp": None, "_derived": None,
                        "theory_params": None, "last": 0}
                       for _ in range(self.n_states)]
        # Timing
        self.timing = timing
        self.n = 0
        self.time_avg = 0

    # What you *must* implement to create your own likelihood:

    # Optional
    def initialize(self):
        """
        Initializes the specifics of this likelihood.
        Note that at this point we know `the `self.input_params``
        and the ``self.output_params``.
        """
        pass

    # Optional
    def add_theory(self):
        """Performs any necessary initialization on the theory side,
        e.g. requests observables"""
        pass

    # Mandatory
    def logp(self, **params_values):
        """
        Computes and returns the log likelihood value.
        Takes as keyword arguments the parameter values.
        To get the derived parameters, pass a `_derived` keyword with an empty dictionary.
        """
        return None

    # Optional
    def close(self):
        """Finalizes the likelihood, if something needs to be done."""
        pass

    # What you *can* implement to create your own likelihood:

    def marginal(self, directions=None, params_values=None):
        """
        (For analytic likelihoods only.)
        Computes the marginal likelihood.
        If nothing is specified, returns the total marginal likelihood.
        If some directions are specified (as a list, tuple or array), returns the marginal
        likelihood pdf over those directions evaluated at the given parameter values.
        """
        self.log.error("Exact marginal likelihood not defined.")
        raise HandledException

    # Other general methods

    def _logp_cached(self, theory_params=None, cached=True, _derived=None, **params_values):
        """
        Wrapper for the `logp` method that caches logp's and derived params.
        If the theory products have been re-computed, re-computes the likelihood anyway.
        """
        params_values = deepcopy(params_values)
        self.log.debug("Got parameters %r", params_values)
        lasts = [self.states[i]["last"] for i in range(self.n_states)]
        try:
            if not cached:
                raise StopIteration
            # Are the parameter values there already?
            i_state = next(i for i in range(self.n_states)
                           if self.states[i]["params"] == params_values)
            # StopIteration not raised, so state exists, but maybe the theory params have
            # changed? In that case, I would still have to re-compute the likelihood
            if self.states[i_state]["theory_params"] != theory_params:
                self.log.debug("Recomputing logp because theory params changed.")
                raise StopIteration
            if _derived is not None:
                _derived.update(self.states[i_state]["derived"] or {})
            self.log.debug("Re-using computed results.")
        except StopIteration:
            # update the (first) oldest one and compute
            i_state = lasts.index(min(lasts))
            self.states[i_state]["params"] = params_values
            self.states[i_state]["theory_params"] = deepcopy(theory_params)
            if self.timing:
                start = time()
            self.states[i_state]["logp"] = self.logp(_derived=_derived, **params_values)
            if self.timing:
                self.n += 1
                self.time_avg = (time() - start + self.time_avg * (self.n - 1)) / self.n
                self.log.debug("Average 'logp' evaluation time: %g s", self.time_avg)
            self.states[i_state]["derived"] = deepcopy(_derived)
        # make this one the current one by decreasing the antiquity of the rest
        for i in range(self.n_states):
            self.states[i]["last"] -= max(lasts)
        self.states[i_state]["last"] = 1
        self.log.debug("Evaluated to logp=%g with derived %r",
                       self.states[i_state]["logp"], self.states[i_state]["derived"])
        return self.states[i_state]["logp"]

    def wait(self):
        if self.delay:
            log.debug("Sleeping for %f seconds.", self.delay)
        sleep(self.delay)

    def d(self):
        return len(self.input_params)

    def __exit__(self, exception_type, exception_value, traceback):
        if self.timing:
            self.log.info("Average 'logp' evaluation time: %g s  (%d evaluations)" %
                          (self.time_avg, self.n))
        self.close()


class LikelihoodExternalFunction(Likelihood):
    def __init__(self, name, info, _theory=None, timing=None):
        self.name = name
        self.log = logging.getLogger(self.name)
        # Load info of the likelihood
        for k in info:
            setattr(self, k, info[k])
        # Store the external function and its arguments
        self.external_function = get_external_function(info[_external], name=self.name)
        argspec = getargspec(self.external_function)
        self.input_params = [p for p in argspec.args if p not in ["_derived", "_theory"]]
        self.has_derived = "_derived" in argspec.args
        if self.has_derived:
            derived_kw_index = argspec.args[-len(argspec.defaults):].index("_derived")
            self.output_params = argspec.defaults[derived_kw_index]
        else:
            self.output_params = []
        self.has_theory = "_theory" in argspec.args
        if self.has_theory:
            theory_kw_index = argspec.args[-len(argspec.defaults):].index("_theory")
            self.needs = argspec.defaults[theory_kw_index]
        # Timing
        self.timing = timing
        self.n = 0
        self.time_avg = 0
        # States, to avoid recomputing
        self.n_states = 3
        self.states = [{"params": None, "logp": None, "derived": None, "last": 0}
                       for _ in range(self.n_states)]

    def add_theory(self):
        if self.has_theory:
            self.theory.needs(**self.needs)

    def logp(self, **params_values):
        # if no derived params defined in external func, delete the "_derived" argument
        if not self.has_derived:
            params_values.pop("_derived")
        if self.has_theory:
            params_values["_theory"] = self.theory
        try:
            return self.external_function(**params_values)
        except:
            self.log.error("".join(
                ["-"] * 16 + ["\n\n"] + list(traceback.format_exception(*sys.exc_info())) +
                ["\n"] + ["-"] * 37))
            self.log.error("The external likelihood '%s' failed at evaluation. "
                      "See traceback on top of this message.", self.name)
            raise HandledException


class LikelihoodCollection(object):
    """
    Likelihood manager:
    Initializes the theory code and the experimental likelihoods.
    """

    def __init__(self, info_likelihood, parameterization, info_theory=None, modules=None,
                 timing=None):
        self.log = logging.getLogger("Likelihood")
        # *IF* there is a theory code, initialize it
        if info_theory:
            name = list(info_theory)[0]
            # If it has an "external" key, wrap it up. Else, load it up
            if _external in info_theory[name]:
                theory_class = info_theory[name][_external]
            else:
                theory_class = get_class(name, kind=_theory)
            self.theory = theory_class(info_theory[name], modules=modules, timing=timing)
        else:
            self.theory = None
        # Initialize individual Likelihoods
        self._likelihoods = odict()
        for name, info in info_likelihood.items():
            # If it has an "external" key, wrap it up. Else, load it up
            if _external in info:
                self._likelihoods[name] = LikelihoodExternalFunction(
                    name, info, _theory=getattr(self, _theory, None), timing=timing)
            else:
                like_class = get_class(name)
                self._likelihoods[name] = like_class(info, modules=modules, timing=timing)
        # Assign input/output parameters
        self._assign_params(parameterization, info_likelihood, info_theory)
        # Do the user-defined post-initialisation, and assign the theory code
        if self.theory:
            self.theory.initialize()
        for like in self:
            self[like].initialize()
            self[like].theory = self.theory
            self[like].add_theory()
        # Store the input params and likelihoods on which each sampled params depends.
        self.sampled_input_dependence = parameterization.sampled_input_dependence()
        self.sampled_like_dependence = odict(
            [[p, [like for like in list(self) + ([_theory] if self.theory else [])
                  if any([(i in self[like].input_params) for i in (i_s or [p])])]]
             for p, i_s in self.sampled_input_dependence.items()])
        # Theory parameters "depend" on every likelihood, since re-computing the theory
        # code forces recomputation of the likelihoods
        for p, ls in self.sampled_like_dependence.items():
            if _theory in ls:
                self.sampled_like_dependence[p] = ([_theory] + list(self._likelihoods))
        # Overhead per likelihood evaluation
        self.overhead = _overhead_per_param * len(parameterization.sampled_params())

    # "get" and iteration operate over the dictionary of likelihoods
    # notice that "get" can get "theory", but the iterator does not!
    def __getitem__(self, key):
        try:
            return self._likelihoods.__getitem__(key) if key != _theory else self.theory
        except KeyError:
            self.log.error("Likelihood '%r' not known", key)
            raise HandledException

    def __iter__(self):
        return self._likelihoods.__iter__()

    def logps(self, input_params, _derived=None, cached=True):
        """
        Computes observables and returns the (log) likelihoods *separately*.
        It takes a list of **input** parameter values, in the same order as they appear
        in the `OrderedDictionary` of the :class:`LikelihoodCollection`.
        To compute the derived parameters, it takes an optional keyword `_derived` as an
        empty list, which is then populated with the derived parameter values.
        """
        self.log.debug("Got input parameters: %r", input_params)
        # Prepare the likelihood-defined derived parameters (only computed if requested)
        # Notice that they are in general different from the sampler-defined ones.
        derived_dict = {}
        # If theory code present, compute the necessary products
        if self.theory:
            theory_params_dict = {p: input_params[p] for p in self.theory.input_params}
            theory_success = self.theory.compute(
                _derived=(derived_dict if _derived is not None else None), cached=cached,
                **theory_params_dict)
            if not theory_success:
                self.log.debug(
                    "Theory code computation failed. Not computing likelihood.")
                if _derived is not None:
                    _derived += [np.nan] * len(self.output_params)
                return np.array([-np.inf for _ in self])
        # Compute each log-likelihood, and optionally get the respective derived params
        logps = []
        for like in self:
            this_params_dict = {p: input_params[p] for p in self[like].input_params}
            this_derived_dict = {} if _derived is not None else None
            logps += [self[like]._logp_cached(
                theory_params=(theory_params_dict if self.theory else None),
                _derived=this_derived_dict, cached=cached, **this_params_dict)]
            derived_dict.update(this_derived_dict or {})
            if _derived is not None:
                derived_dict[_chi2 + _separator + like] = -2*logps[-1]
        # Turn the derived params dict into a list and return
        if _derived is not None:
            _derived += [derived_dict[p] for p in self.output_params]
        return np.array(logps)

    def logp(self, input_params, _derived=None, cached=True):
        """
        Computes observables and returns the (log) likelihood.
        It takes a list of **sampled** parameter values, in the same order as they appear
        in the `OrderedDictionary` of the :class:`LikelihoodCollection`.
        To compute the derived parameters, it takes an optional keyword `_derived` as an
        empty list, which is then populated with the derived parameter values.
        """
        return np.sum(self.logps(input_params, _derived=_derived))

    def marginal(self, *args):
        self.log.error("Marginal not implemented for >1 likelihoods. Sorry!")
        raise HandledException

    def _assign_params(self, parameterization, info_likelihood, info_theory=None):
        """
        Assign parameters to likelihoods, following the algorithm explained in
        :doc:`DEVEL`.
        """
        self.input_params = list(parameterization.input_params())
        self.output_params = list(parameterization.output_params())
        params_assign = odict([
            ["input", odict([[p, []] for p in self.input_params])],
            ["output", odict([[p, []] for p in self.output_params])]])
        agnostic_likes = {"input": [], "output": []}
        for kind, option, prefix in (
                ["input", _input_params, _input_params_prefix],
                ["output", _output_params, _output_params_prefix]):
            for like in (list(self) + ([_theory] if self.theory else [])):
                # "one" only takes leftover parameters
                if like == "one":
                    continue
                # Identify parameters understood by this likelihood/therory
                # 1a. Does it have input/output params list?
                #     (takes into account that for callables, we can ignore elements)
                if getattr(self[like], option, None) is not None:
                    for p in getattr(self[like], option):
                        try:
                            params_assign[kind][p] += [like]
                        except KeyError:
                            if kind == "input":
                                # If external function, no problem: it may have default value
                                if not isinstance(self[like], LikelihoodExternalFunction):
                                    self.log.error("Parameter '%s' needed as input for '%s', "
                                                   "but not provided.", p, like)
                                    raise HandledException
                # 2. Is there a params prefix?
                elif getattr(self[like], prefix, None) is not None:
                    for p in params_assign[kind]:
                        if p.startswith(getattr(self[like], prefix)):
                            params_assign[kind][p] += [like]
                # 3. Does it have a general (mixed) list of params?
                elif getattr(self[like], _params, None) is not None:
                    for p in getattr(self[like], _params):
                        if p in params_assign[kind]:
                            params_assign[kind][p] += [like]
                # 4. No parameter knowledge: store as parameter agnostic
                else:
                    agnostic_likes[kind] += [like]
                # Check that there is only one non-knowledgeable element, and assign unused params
                if len(agnostic_likes[kind]) > 1:
                    self.log.error("More than once parameter-agnostic likelihood/theory "
                                   "with respect to %s parameters: %r. Cannot decide "
                                   "parameter assignements.", kind, param_agnostic_likes)
                    raise HandledException
                elif agnostic_likes[kind]:  # if there is only one
                    for p, assigned in params_assign[kind].items():
                        if not assigned:
                            params_assign[kind][p] = [agnostic_likes[kind][0]]
        # Check that, if a theory code is present, it does not share input parameters with
        # any likelihood (because of the theory+experimental model separation)
        if self.theory:
            for p, assigned in params_assign["input"].items():
                 if _theory in assigned and len(assigned) > 1:
                     self.log.error("Some parameter has been asigned to the theory code "
                                    "AND a likelihood, and that is not allowed: {%s: %r}",
                                    p, assigned)
                     raise HandledException
        # If unit likelihood is present, assign all *non-theory* inputs to it
        if "one" in self:
            for p, assigned in params_assign["input"].items():
                if _theory not in assigned:
                    assigned += ["one"]
        # If there are unassigned input params, fail
        unassigned_input = [
            p for p, assigned in params_assign["input"].items() if not assigned]
        if unassigned_input:
            self.log.error("Could not find whom to assign input parameters %r.",
                           unassigned_input)
            raise HandledException
        # Assign the "chi2__" output parameters
        for p in params_assign["output"]:
            if p.startswith(_chi2 + _separator):
                like = p[len(_chi2 + _separator):]
                if like not in list(self):
                    self.log.error("Your derived parameters depend on an unknown "
                                   "likelihood: '%s'", like)
                    raise HandledException
                # They may have been already assigned to an agnostic likelihood,
                # so purge first: no "=+"
                params_assign["output"][p] = [like]
        # Check that output parameters are assigned exactly once
        unassigned_output = [
            p for p, assigned in params_assign["output"].items() if not assigned]
        multiassigned_output = {
            p: assigned for p, assigned in params_assign["output"].items()
            if len(assigned) > 1}
        if unassigned_output:
            self.log.error("Could not find whom to assign output parameters %r.",
                           unassigned_output)
            raise HandledException
        if multiassigned_output:
            self.log.error("Output params can only be computed by one likelihood/theory, "
                           "but some were claimed by more than one: %r.",
                           multiassigned_output)
            raise HandledException
        # Finished! Assign and update infos
        params_assign_inv = odict([["input", odict()], ["output", odict()]])
        for kind, option, attr in (
                ["input", _input_params, "input_params"],
                ["output", _output_params, "output_params"]):
            for like in list(self) + ([_theory] if self.theory else []):
                setattr(
                    self[like], attr,
                    [p for p, assign in params_assign[kind].items() if like in assign])
                # Update infos!
                if like != _theory:
                    info_likelihood[like].pop(_params, None)
                    info_likelihood[like][option] = getattr(self[like], attr)
                elif self.theory:
                    name = list(info_theory)[0]
                    info_theory[name].pop(_params, None)
                    info_theory[name][option] = getattr(self[like], attr)
        if self.log.getEffectiveLevel() <= logging.DEBUG:
            self.log.debug("Parameters were assigned as follows:")
            for like in list(self) + ([_theory] if self.theory else []):
                self.log.debug("- %r:", like)
                self.log.debug("     Input:  %r", self[like].input_params)
                self.log.debug("     Output: %r", self[like].output_params)

    def _speeds_of_params(self, int_speeds=False, fast_slow=False):
        """
        Separates the sampled parameters in blocks according to the likelihood (or theory)
        re-evaluation that changing each one of them involves. Using the appoximate speed
        (i.e. inverse evaluation time in seconds) of each likelihood, sorts the blocks in
        an optimal way, in ascending order of speed *per full block iteration*.

        Returns tuples of ``(speeds), (params_in_block)``, sorted by ascending speeds,
        where speeds are *per param* (though optimal blocking is chosen by speed
        *per full block*).

        If ``int_speeds=True``, returns integer speeds, instead of speeds in 1/s.

        If ``fast_slow=True``, returns just 2 blocks: a fast and a slow one, each one
        assigned its slowest per-parameter speed.
        """
        # Fill unknown speeds with the value of the slowest one, and clip with overhead
        speeds = np.array([getattr(self[like], "speed", -1) for like in self] +
                          ([getattr(self.theory, "speed", -1)] if self.theory else []),
                          dtype=float)
        # Add overhead to the defined ones, and clip to the slowest the undefined ones
        speeds[speeds > 0] = (speeds[speeds > 0] ** -1 + self.overhead) ** -1
        try:
            speeds = np.clip(speeds, min(speeds[speeds > 0]), None)
        except ValueError:
            # No speeds specified
            speeds = np.ones(len(speeds))
        likes = list(self) + ([_theory] if self.theory else [])
        for i, like in enumerate(likes):
            self[like].speed = speeds[i]
        # Compute "footprint"
        # i.e. likelihoods (and theory) that we must recompute when each parameter changes
        footprints = np.zeros((len(self.sampled_like_dependence), len(likes)), dtype=int)
        for i, ls in enumerate(self.sampled_like_dependence.values()):
            for j, like in enumerate(likes):
                footprints[i, j] = like in ls
        # Group parameters by footprint
        different_footprints = list(set([tuple(row) for row in footprints.tolist()]))
        blocks = [[p for ip, p in enumerate(self.sampled_like_dependence)
                   if all(footprints[ip] == fp)] for fp in different_footprints]
        # Find optimal ordering, such that one minimises the time it takes to vary every
        # parameter, one by one, in a basis in which they are mixed-down (i.e after a
        # Cholesky transformation)
        # To do that, compute that "total cost" for every permutation of the blocks order,
        # and find the minumum.
        n_params_per_block = np.array([len(b) for b in blocks])
        self._costs = 1 / np.array(speeds)
        self._footprints = np.array(different_footprints)
        self._lower = np.tri(len(n_params_per_block))

        def get_cost_per_param_per_block(ordering):
            """
            Computes cumulative cost per parameter for each block, given ordering.
            """
            footprints_chol = np.minimum(
                1, self._footprints[ordering].T.dot(self._lower).T)
            return footprints_chol.dot(self._costs)

        orderings = list(permutations(np.arange(len(n_params_per_block))))
        costs_per_param_per_block = np.array(
            [get_cost_per_param_per_block(list(o)) for o in orderings])
        total_costs = np.array(
            [n_params_per_block[list(o)].dot(costs_per_param_per_block[i])
             for i, o in enumerate(orderings)])
        i_optimal = np.argmin(total_costs)
        optimal_ordering = orderings[i_optimal]
        blocks = [blocks[i] for i in optimal_ordering]
        costs_per_param_per_block = costs_per_param_per_block[i_optimal]
        # This costs are *cumulative-down* (i.e. take into account the cost of varying the
        # parameters below the present one). Subtract that effect so that its inverse,
        # the speeds, are equivalent to oversampling factors
        costs_per_param_per_block[:-1] -= costs_per_param_per_block[1:]
        params_speeds = 1 / costs_per_param_per_block
        if int_speeds:
            # Oversampling precision: smallest difference in oversampling to be ignored.
            speed_precision = 1 / 10
            speeds = np.array(np.round(np.array(
                params_speeds) / min(params_speeds) / speed_precision), dtype=int)
            params_speeds = np.array(
                speeds / np.ufunc.reduce(np.frompyfunc(gcd, 2, 1), speeds), dtype=int)
        self.log.debug("Optimal ordering of parameter blocks: %r with speeds %r",
                       blocks, params_speeds)
        # Fast-slow separation: chooses separation that maximizes log-difference in speed
        # (speed per parameter in a combination of blocks is the slowest one)
        if fast_slow:
            if len(blocks) > 1:
                log_differences = np.zeros(len(blocks) - 1)
                for i in range(len(blocks) - 1):
                    log_differences[i] = (np.log(np.min(params_speeds[:i + 1])) -
                                          np.log(np.min(params_speeds[i + 1:])))
                i_max = np.argmin(log_differences)
                blocks = (
                    lambda l: [list(chain(*l[:i_max + 1])), list(chain(*l[i_max + 1:]))])(blocks)
                # In this case, speeds must be *cumulative*, since I am squashing blocks
                cum_inv = lambda ss: 1 / (sum(1 / ss))
                params_speeds = (
                    lambda l: [cum_inv(l[:i_max + 1]), cum_inv(l[i_max + 1:])])(params_speeds)
                self.log.debug("Fast-slow blocking: %r with speeds %r",
                               blocks, params_speeds)
            else:
                self.log.warning("Requested fast/slow separation, "
                                 "but all pararameters have the same speed.")
        return params_speeds, blocks

    def _check_speeds_of_params(self, blocking):
        """
        Checks the correct formatting of the given parameter blocking.

        `blocking` is a list of tuples `(speed, (param1, param2, etc))`.

        Returns tuples of ``(speeds), (params_in_block)``.
        """
        try:
            speeds, blocks = zip(*list(blocking))
            speeds = np.array(speeds)
        except:
            raise HandledException(
                "Manual blocking not understood. Check documentation.")
        sampled_params = list(self.sampled_like_dependence)
        check = compare_params_lists(
            list(chain(*blocks)), sampled_params)
        duplicate = check.pop("duplicate_A", None)
        missing = check.pop("B_but_not_A", None)
        unknown = check.pop("A_but_not_B", None)
        if duplicate:
            self.log.error("Manual blocking: repeated parameters: %r", duplicate)
            raise HandledException
        if missing:
            self.log.error("Manual blocking: missing parameters: %r", missing)
            raise HandledException
        if unknown:
            self.log.error("Manual blocking: unkown parameters: %r", unknown)
            raise HandledException
        if (speeds != np.sort(speeds)).all():
            self.log.warn("Manual blocking: speed-blocking *apparently* non-optimal: "
                          "sort by ascending speed when possible")
        return speeds, blocks

    def _get_auto_covmat(self, params_info):
        """
        Tries to get an automatic covariance matrix for the current model and data.

        ``params_info`` should contain preferably the slow parameters only.
        """
        likes_renames = {like: {_p_renames: getattr(self[like], _p_renames, [])}
                         for like in self if like != _theory}
        try:
            return self.theory.get_auto_covmat(params_info, likes_renames)
        except:
            return None

    def dump_timing(self):
        avg_times_evals = odict([
            [like, {"t": self[like].time_avg, "n": self[like].n}] for like in
            (([_theory] if self.theory else []) + [like for like in self])
            if getattr(self[like], _timing)])
        if avg_times_evals:
            sep = "\n   "
            self.log.info(
                "Average computation time:" + sep + sep.join(
                ["%s : %g s (%d evaluations)" % (name, vals["t"], vals["n"])
                 for name, vals in avg_times_evals.items()]))

    # Python magic for the "with" statement
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if self.theory:
            self.theory.__exit__(exception_type, exception_value, traceback)
        for like in self:
            self[like].__exit__(exception_type, exception_value, traceback)



"""
.. module:: model

:Synopsis: Wrapper for models: parameterization+prior+likelihood+theory
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import numpy as np
from copy import deepcopy
from collections import namedtuple, OrderedDict as odict

# Local
from cobaya.conventions import _likelihood, _prior, _params, _theory, _timing
from cobaya.conventions import _path_install, _debug, _debug_default, _debug_file
from cobaya.conventions import _input_params, _output_params
from cobaya.input import get_full_info
from cobaya.parameterization import Parameterization
from cobaya.prior import Prior
from cobaya.likelihood import LikelihoodCollection as Likelihood
from cobaya.log import HandledException, logger_setup
from cobaya.yaml import yaml_dump

# Logger
import logging

# Log-posterior namedtuple
logposterior = namedtuple("logposterior", ["logpost", "logpriors", "loglikes", "derived"])
logposterior.__new__.__defaults__ = (None, None, [], [])


def get_model(info):
    assert hasattr(info, "keys"), (
        "The first argument must be a dictionary with the info needed for the model. "
        "If you were trying to pass the name of an input file instead, "
        "load it first with 'cobaya.input.load_input', "
        "or, if you were passing a yaml string, load it with 'cobaya.yaml.yaml_load'.")
    # Configure the logger ASAP
    # Just a dummy import before configuring the logger, until I fix root/individual level
    import getdist
    logger_setup(info.pop(_debug, _debug_default), info.pop(_debug_file, None))
    # Create the full input information, including defaults for each module.
    info = deepcopy(info)
    ignored_info = {}
    for k in list(info):
        if k not in [_params, _likelihood, _prior, _theory, _path_install, _timing]:
            ignored_info.update({k: info.pop(k)})
    import logging
    if ignored_info:
        logging.getLogger(__name__.split(".")[-1]).warn(
            "Ignored blocks/options: %r", list(ignored_info))
    full_info = get_full_info(info)
    if logging.root.getEffectiveLevel() <= logging.DEBUG:
        logging.getLogger(__name__.split(".")[-1]).debug(
            "Input info updated with defaults (dumped to YAML):\n%s",
            yaml_dump(full_info))
    # Initialize the posterior and the sampler
    return Model(full_info[_params], full_info[_likelihood], full_info.get(_prior),
                 full_info.get(_theory), modules=info.get(_path_install),
                 timing=full_info.get(_timing))


class Model(object):
    """
    Class containing all the information necessary to compute the unnormalized posterior.

    Allows for low-level interaction with the theory code, prior and likelihood.

    **NB:** do not initialize this class directly; use :func:`~model.get_model` instead,
    with some info as input.
    """

    def __init__(self, info_params, info_likelihood, info_prior=None, info_theory=None,
                 modules=None, timing=None, allow_renames=True):
        self.log = logging.getLogger(self.__class__.__name__)
        self._full_info = {
            _params: deepcopy(info_params),
            _likelihood: deepcopy(info_likelihood)}
        if not self._full_info[_likelihood]:
            self.log.error("No likelihood requested!")
            raise HandledException
        for k, v in ((_prior, info_prior), (_theory, info_theory),
                     (_path_install, modules), (_timing, timing)):
            if v not in (None, {}):
                self._full_info[k] = deepcopy(v)
        self.parameterization = Parameterization(
            self._full_info[_params], allow_renames=allow_renames)
        self.prior = Prior(self.parameterization, self._full_info.get(_prior, None))
        self.likelihood = Likelihood(
            self._full_info[_likelihood], self.parameterization,
            self._full_info.get(_theory), modules=modules,timing=timing)

    def info(self):
        """
        Returns a copy of the information used to create the model, including defaults.
        """
        return deepcopy(self._full_info)

    def _to_sampled_array(self, params_values):
        """
        Internal method to interact with the prior.
        Needs correct (not renamed) parameter names.
        """
        if hasattr(params_values, "keys"):
            params_values_array = np.array(list(params_values.values()))
        else:
            params_values_array = np.atleast_1d(params_values)
            if params_values_array.shape[0] != self.prior.d():
                self.log.error("Wrong dimensionality: it's %d and it should be %d.",
                               len(params_values_array), self.prior.d())
                raise HandledException
        if len(params_values_array.shape) >= 2:
            self.log.error("Cannot take arrays of points as inputs, just single points.")
            raise HandledException
        return params_values_array

    def logpriors(self, params_values, make_finite=False):
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns the log-values of the priors, in the same order as it is returned by
        ``list([your_model].prior)``. The first one, named ``0``, corresponds to the
        product of the 1-dimensional priors specified in the ``params`` block, and it's
        normalized (in general, the external prior densities aren't).

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.
        """
        if hasattr(params_values, "keys"):
            params_values = self.parameterization._check_sampled(**params_values)
        params_values_array = self._to_sampled_array(params_values)
        logpriors = self.prior.logps(params_values_array)
        if make_finite:
            return np.nan_to_num(logpriors)
        return logpriors

    def logprior(self, params_values, make_finite=False):
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns the log-value of the prior (in general, unnormalized, unless the only
        priors specified are the 1-dimensional ones in the ``params`` block).

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.
        """
        logprior = np.sum(self.logpriors(params_values))
        if make_finite:
            return np.nan_to_num(logprior)
        return logprior

    def loglikes(self, params_values, return_derived=True, make_finite=False, cached=True,
                 _no_check=False):
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns a tuple ``(loglikes, derived_params)``, where ``loglikes`` are the
        log-values of the likelihoods (unnormalized, in general) in the same order as
        it is returned by ``list([your_model].likelihood)``, and ``derived_params``
        are the values of the derived parameters in the order given by
        ``list([your_model].parameterization.derived_params())``.

        To return just the list of log-likelihood values, make ``return_derived=False``.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        if hasattr(params_values, "keys") and not _no_check:
            params_values = self.parameterization._check_sampled(**params_values)
        _derived = [] if return_derived else None
        loglikes = self.likelihood.logps(
            self.parameterization._to_input(params_values),
            _derived=_derived, cached=cached)
        if make_finite:
            loglikes = np.nan_to_num(loglikes)
        if return_derived:
            derived_sampler = self.parameterization._to_derived(_derived)
            if self.log.getEffectiveLevel() <= logging.DEBUG:
                self.log.debug(
                    "Computed derived parameters: %s",
                    dict(zip(self.parameterization.derived_params(), derived_sampler)))
            return loglikes, derived_sampler
        return loglikes

    def loglike(self, params_values, return_derived=True, make_finite=False, cached=True):
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns a tuple ``(loglike, derived_params)``, where ``loglike`` is the log-value
        of the likelihood (unnormalized, in general), and ``derived_params``
        are the values of the derived parameters in the order given by
        ``list([your_model].parameterization.derived_params())``.
        If the model contains multiple likelihoods, the sum of the loglikes is returned.

        To return just the list of log-likelihood values, make ``return_derived=False``.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        ret_value = self.loglikes(
            params_values, return_derived=return_derived, cached=cached)
        if return_derived:
            loglike = np.sum(ret_value[0])
            if make_finite:
                return np.nan_to_num(loglike), ret_value[1]
            return loglike, ret_value[1]
        else:
            loglike = np.sum(ret_value)
            if make_finite:
                return np.nan_to_num(loglike)
            return loglike

    def logposterior(
            self, params_values, return_derived=True, make_finite=False, cached=True):
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns the a ``logposterior`` ``NamedTuple``, with the following fields:

        - ``logpost``: log-value of the posterior.
        - ``logpriors``: log-values of the priors, in the same order as in
          ``list([your_model].prior)``. The first one, corresponds to the
          product of the 1-dimensional priors specified in the ``params``
          block. Except for the first one, the priors are unnormalized.
        - ``loglikes``: log-values of the likelihoods (unnormalized, in general),
          in the same order as in ``list([your_model].likelihood)``.
        - ``derived``: values of the derived parameters in the order given by
          ``list([your_model].parameterization.derived_params())``.

        Only computes the log-likelihood and the derived parameters if the prior is
        non-null (otherwise the fields ``loglikes`` and ``derived`` are empty lists).

        To ignore the derived parameters, make ``return_derived=False``.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        if hasattr(params_values, "keys"):
            params_values = self.parameterization._check_sampled(**params_values)
        params_values_array = self._to_sampled_array(params_values)
        if self.log.getEffectiveLevel() <= logging.DEBUG:
            self.log.debug(
                "Posterior to be computed for parameters %s",
                dict(zip(self.parameterization.sampled_params(), params_values_array)))
        if not np.all(np.isfinite(params_values_array)):
            self.log.error(
                "Got non-finite parameter values: %r",
                dict(zip(self.parameterization.sampled_params(), params_values_array)))
            raise HandledException
        # Notice that we don't use the make_finite in the prior call,
        # to correctly check if we have to compute the likelihood
        logpriors = self.logpriors(params_values_array, make_finite=False)
        logpost = sum(logpriors)
        if -np.inf not in logpriors:
            l = self.loglikes(params_values, return_derived=return_derived,
                              make_finite=make_finite, cached=cached, _no_check=True)
            loglikes, derived_sampler = l if return_derived else (l, [])
            logpost += sum(loglikes)
        else:
            loglikes = []
            derived_sampler = []
        if make_finite:
            logpriors = np.nan_to_num(logpriors)
            logpost = np.nan_to_num(logpost)
        return logposterior(logpost=logpost, logpriors=logpriors,
                            loglikes=loglikes, derived=derived_sampler)

    def logpost(self, params_values, make_finite=False, cached=True):
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns the log-value of the posterior.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        return self.logposterior(params_values, make_finite=make_finite,
                                 return_derived=False, cached=cached)[0]

    def dump_timing(self):
        """
        Prints the average computation time of the theory code and likelihoods.

        It's more reliable the more times the likelihood has been evaluated.
        """
        self.likelihood.dump_timing()

    # Python magic for the "with" statement
    def __enter__(self):
        return self

    def __exit__(self, exception_type=None, exception_value=None, traceback=None):
        self.likelihood.__exit__(exception_type, exception_value, traceback)

    def close(self):
        self.__exit__()



"""
.. module:: samplers.minimize

:Synopsis: Posterior/likelihood *maximization* (i.e. chi^2 minimization).
:Author: Jesus Torrado (though it's just a wrapper of ``scipy.optimize.minimize``)

This is a **maximizator** for posteriors or likelihoods, using
`scipy.optimize.minimize <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.

It is pretty self-explanatory: just look at the comments on the defaults below.

It is recommended to run a couple of parallel MPI processes:
it will finally pick the best among the results.
"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import numpy as np
from scipy.optimize import minimize as scpminimize
import logging

# Local
from cobaya.sampler import Sampler
from cobaya.mpi import get_mpi_rank, get_mpi_size, get_mpi_comm
from cobaya.collection import OnePoint
from cobaya.log import HandledException


class minimize(Sampler):
    def initialize(self):
        """Prepares the arguments for `scipy.minimize`."""
        if not get_mpi_rank():
            self.log.info("Initializing")
        self.logp = ((lambda x: self.model.logposterior(x, make_finite=True)[0])
                     if not self.ignore_prior else
                     (lambda x: sum(self.model.loglikes(x, return_derived=True)[0])))
        # Initial point: sample from reference and make sure that it has finite lik/post
        this_logp = -np.inf
        while not np.isfinite(this_logp):
            initial_point = self.model.prior.reference()
            this_logp = self.logp(initial_point)
        self.kwargs = {
            "fun": (lambda x: -self.logp(x)),
            "x0": initial_point,
            "bounds": self.model.prior.bounds(confidence_for_unbounded=0.999),
            "tol": self.tol,
            "options": {
                "maxiter": self.maxiter,
                "disp": (self.log.getEffectiveLevel() == logging.DEBUG)}}
        self.kwargs.update(self.override or {})
        self.log.debug("Arguments for scipy.optimize.minimize:\n%r", self.kwargs)

    def run(self):
        """
        Runs `scipy.minimize`
        """
        self.log.info("Starting minimization.")
        self.result = scpminimize(**self.kwargs)
        if self.result.success:
            self.log.info("Finished succesfully.")
        else:
            self.log.error("Finished Unsuccesfully.")

    def close(self, *args):
        """
        Determines success (or not), chooses best (if MPI)
        and produces output (if requested).
        """
        # If something failed
        if not hasattr(self, "result"):
            return
        if get_mpi_size():
            results = get_mpi_comm().gather(self.result, root=0)
            if not get_mpi_rank():
                self.result = results[np.argmin([r.fun for r in results])]
        if not get_mpi_rank():
            if not self.result.success:
                self.log.error("Maximization failed! Here is the `scipy` raw result:\n%r",
                               self.result)
                raise HandledException
            self.log.info("log%s maximized at %g",
                          "likelihood" if self.ignore_prior else "posterior",
                          -self.result.fun)
            post = self.model.logposterior(self.result.x)
            recomputed_max = sum(post.loglikes) if self.ignore_prior else post.logpost
            if not np.allclose(-self.result.fun, recomputed_max):
                self.log.error("Cannot reproduce result. Something bad happened. "
                               "Recomputed max: %g at %r", recomputed_max, self.result.x)
                raise HandledException
            self.maximum = OnePoint(
                self.model, self.output, name="maximum",
                extension=("likelihood" if self.ignore_prior else "posterior"))
            self.maximum.add(self.result.x, derived=post.derived, logpost=post.logpost,
                             logpriors=post.logpriors, loglikes=post.loglikes)
            self.log.info("Parameter values at maximum:\n%s"%self.maximum.data.to_string())
            self.maximum._out_update()

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.

        Returns:
           The :class:`OnePoint` that maximizes the posterior or likelihood (depending on
           ``ignore_prior``), and the `scipy.optimize.OptimizeResult
           <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html>`_
           instance.
        """
        if not get_mpi_rank():
            return {"maximum": self.maximum, "OptimizeResult": self.result}



"""
.. module:: likelihoods._H0_prototype

:Synopsis: Prototype class for local Hubble parameter measurements
:Author: Jesus Torrado

This is a simple gaussian likelihood for the latest local :math:`H_0` measurements
using a combination of different data.

It defines the following likelihoods

- ``H0_riess2018a``: A legacy local measurement of :math:`H_0`, used in the analysis of
  Planck data for the Planck 2018 release.
- ``H0_riess2018b``: The latest local measurement of :math:`H_0`.

.. note::

   If you use this likelihood, please cite it as:

   - ``H0_riess2018a``: A. Riess et al, *"New Parallaxes of Galactic Cepheids from
     Spatially Scanning the Hubble Space Telescope: Implications for the Hubble Constant"*
     `(arXiv:1801.01120) <https://arxiv.org/abs/1801.01120>`_

   - ``H0_riess2018b``: A. Riess et al, *"Milky Way Cepheid Standards for Measuring Cosmic
     Distances and Application to Gaia DR2: Implications for the Hubble Constant"*
     `(arXiv:1804.10655) <https://arxiv.org/abs/1804.10655>`_


Using a different measurement
-----------------------------

If you would like to use different values for the :math:`H_0` constraint, as a mean and a
standard deviation, simply add the following likelihood, substituting ``mu_H0`` and
``sigma_H0`` for the respective values:

.. code:: yaml

   likelihood:
     my_H0: 'lambda _theory={"H0": None}: stats.norm.logpdf(_theory.get_param("H0"), loc=mu_H0, scale=sigma_H0)'

"""

# Python 2/3 compatibility
from __future__ import division

# Global
from scipy.stats import norm

# Local
from cobaya.likelihood import Likelihood


class _H0_prototype(Likelihood):

    def initialize(self):
        self.norm = norm(loc=self.H0, scale=self.H0_std)

    def add_theory(self):
        self.theory.needs(H0=None)

    def logp(self, **params_values):
        H0_theory = self.theory.get_param("H0")
        return self.norm.logpdf(H0_theory)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from itertools import chain
# Package data and conventions
from cobaya import __author__, __version__, __url__
from cobaya.conventions import subfolders


# Get the long description from the README file
def get_long_description():
    with open(path.join(path.abspath(path.dirname(__file__)), 'README.rst'),
              encoding='utf-8') as f:
        lines = f.readlines()
        i = -1
        while '=====' not in lines[i]:
            i -= 1
        return "".join(lines[:i])


setup(
    name='cobaya',
    version=__version__,
    description='Bayesian Analysis in Cosmology',
    long_description=get_long_description(),
    url=__url__,
    author=__author__,
    license='LGPL',
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'License :: OSI Approved :: '
            'GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*',
    keywords='montecarlo sampling cosmology',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=['numpy>=1.12.0', 'scipy >= 1.0', 'pandas>=0.20',
                      'PyYAML>=5.1', 'wget>=3.2', 'imageio>=2.2.0',
                      'GetDist>=0.3.0', 'fuzzywuzzy>=0.17'],
    extras_require={
        'test': ['pytest', 'pytest-xdist', 'flaky', 'mpi4py']},
    package_data={
        'cobaya': list(chain(*[['%s/*/*.yaml' % folder, '%s/*/*.bibtex' % folder]
                               for folder in subfolders.values()]))},
    entry_points={
        'console_scripts': [
            'cobaya-install=cobaya.install:install_script',
            'cobaya-create-image=cobaya.containers:create_image_script',
            'cobaya-prepare-data=cobaya.containers:prepare_data_script',
            'cobaya-run=cobaya.run:run_script',
            'cobaya-citation=cobaya.citation:citation_script',
            'cobaya-grid-create=cobaya.grid_tools:MakeGridScript',
            'cobaya-grid-run=cobaya.grid_tools.runbatch:run',
            'cobaya-cosmo-generator=cobaya.cosmo_input:gui_script',
        ],
    },
)



# Python 2/3 compatibility
from __future__ import absolute_import, division

# Global
import os
from random import choice
from itertools import chain
from collections import OrderedDict as odict
from six import string_types
import numpy as np

# Local
from cobaya.yaml import yaml_load_file, yaml_dump_file
from cobaya.conventions import _covmats_file, _p_renames, _path_install

# Logger
import logging

log = logging.getLogger(__name__.split(".")[-1])

covmat_folders = ["{%s}/data/planck_supp_data_and_covmats/covmats/" % _path_install,
                  "{%s}/data/bicep_keck_2015/BK15_cosmomc/planck_covmats/" % _path_install]


def get_covmat_database(modules, cached=True):
    # Get folders with corresponding modules installed
    installed_folders = [folder for folder in covmat_folders
                         if os.path.exists(folder.format(**{_path_install: modules}))]
    covmats_database_fullpath = os.path.join(modules, _covmats_file)
    # Check if there is a usable cached one
    if cached:
        try:
            covmat_database = yaml_load_file(covmats_database_fullpath)
            assert set(covmat_database) == set(installed_folders)
            return covmat_database
        except:
            log.info("No cached covmat database present, not usable or not up-to-date. "
                     "Will be re-created and cached.")
            pass
    # Create it (again)
    covmat_database = odict()
    for folder in installed_folders:
        covmat_database[folder] = []
        folder_full = folder.format(**{_path_install: modules}).replace("/", os.sep)
        for filename in os.listdir(folder_full):
            try:
                with open(os.path.join(folder_full, filename)) as covmat:
                    header = covmat.readline()
                assert header.strip().startswith("#")
                params = header.strip().lstrip("#").split()
            except:
                continue
            covmat_database[folder].append({"name": filename, "params": params})
    if cached:
        yaml_dump_file(covmats_database_fullpath, covmat_database, error_if_exists=False)
    return covmat_database


def get_best_covmat(modules, slow_params_info, likelihoods_info, cached=True):
    covmats_database = get_covmat_database(modules, cached=cached)
    # Select first based on number of slow parameters
    str_to_list = lambda x: ([x] if isinstance(x, string_types) else x)
    params_renames = set(chain(*[
        [p] + str_to_list(info.get(_p_renames, [])) for p, info in slow_params_info.items()]))
    get_score_params = (
        lambda covmat_params: len(set(covmat_params).intersection(params_renames)))
    highest_score = 0
    best = []
    for folder, covmats in covmats_database.items():
        for covmat in covmats:
            score = get_score_params(covmat["params"])
            if score > highest_score:
                highest_score = score
                best = []
            if score == highest_score:
                best.append({
                    "folder": folder, "name": covmat["name"], "params": covmat["params"]})
    if highest_score == 0:
        log.warning(
            "No covariance matrix found including at least one of the given parameters")
        return None
    # Sub-select by number of likelihoods
    likes_renames = set(chain(*[[like] + str_to_list(info.get(_p_renames, []))
                                for like, info in likelihoods_info.items()]))
    get_score_likes = (
        lambda covmat_name: len([0 for like in likes_renames if like in covmat_name]))
    highest_score = 0
    best_2 = []
    for covmat in best:
        score = get_score_likes(covmat["name"])
        if score > highest_score:
            highest_score = score
            best_2 = []
        if score == highest_score:
            best_2.append(covmat)
    # Finally, in case there is more than one, select shortest #params and name (simpler!)
    # #params first, to avoid extended models with shorter covmat name
    get_score_simpler_params = lambda covmat_params: -len(covmat_params)
    highest_score = -np.inf
    best_3 = []
    for covmat in best_2:
        score = get_score_simpler_params(covmat["params"])
        if score > highest_score:
            highest_score = score
            best_3 = []
        if score == highest_score:
            best_3.append(covmat)
    get_score_simpler_name = (
        lambda covmat_name: -len(covmat_name.replace("_", " ").replace("-", " ").split()))
    highest_score = -np.inf
    best_4 = []
    for covmat in best_3:
        score = get_score_simpler_name(covmat["name"])
        if score > highest_score:
            highest_score = score
            best_4 = []
        if score == highest_score:
            best_4.append(covmat)
    # if there is more than one (unlikely), just pick one at random
    if len(best_4) > 1:
        log.warning("WARNING: >1 possible best covmats: %r" % [b["name"] for b in best_4])
    return best_4[choice(range(len(best_4)))]



"""
.. module:: citation

:Synopsis: Tools and script to get the references to cite
:Author: Jesus Torrado

Inspired by a similar characteristic of
`CosmoSIS <https://bitbucket.org/joezuntz/cosmosis/wiki/Home>`_.

"""
# Python 2/3 compatibility
from __future__ import absolute_import, division, print_function

# Global
import os
from collections import OrderedDict as odict

# Local
from cobaya.tools import get_folder, make_header, warn_deprecation
from cobaya.input import get_modules


def get_citation_info(module, kind):
    folder = get_folder(module, kind, absolute=True)
    filename = os.path.join(folder, module + ".bibtex")
    try:
        with open(filename, "r") as f:
            lines = "".join(f.readlines())
    except IOError:
        if not os.path.isdir(folder):
            lines = "[Module '%s.%s' not known.]" % (kind, module)
        else:
            lines = "[no citation information found]"
    return lines + "\n"


def citation(*infos):
    blocks_text = odict([["Cobaya", "[Paper in preparation]"]])
    for kind, modules in get_modules(*infos).items():
        for module in modules:
            blocks_text["%s:%s"%(kind, module)] = get_citation_info(module, kind)
    return blocks_text

def prettyprint_citation(blocks_text):
    txt = ""
    for block, text in blocks_text.items():
        if not txt.endswith("\n\n"):
            txt += "\n\n"
        txt += make_header(*block.split(":")) + "\n" + text
    return txt.lstrip().rstrip() + "\n"

# Command-line script
def citation_script():
    warn_deprecation()
    from cobaya.mpi import am_single_or_primary_process
    if am_single_or_primary_process():
        warn_deprecation()
        # Configure the logger ASAP
        from cobaya.log import logger_setup
        logger_setup()
        # Parse arguments and launch
        import argparse
        parser = argparse.ArgumentParser(description="Cobaya's citation tool.")
        parser.add_argument("files", action="store", nargs="+", metavar="input_file.yaml",
                            help="One or more input files.")
        from cobaya.input import load_input
        infos = [load_input(f) for f in parser.parse_args().files]
        print(prettyprint_citation(citation(*infos)))



"""
.. module:: planck_2015_lensing_cmblikes

:Synopsis: Alternative version of the Planck lensing likelihood
:Author: Antony Lewis

"""

# Global
from __future__ import division, print_function
import os
import logging

# Local
from cobaya.likelihoods._cmblikes_prototype import _cmblikes_prototype
from cobaya.install import download_github_release


class planck_2015_lensing_cmblikes(_cmblikes_prototype):
    pass


# Installation routines ##################################################################

# name of the supplementary data and covmats repo/folder
supp_data_name = "planck_supp_data_and_covmats"
supp_data_version = "v1.1"


def get_path(path):
    return os.path.realpath(os.path.join(path, "data", supp_data_name))


def is_installed(**kwargs):
    return os.path.exists(os.path.realpath(
        os.path.join(kwargs["path"], "data", supp_data_name)))


def install(path=None, force=False, code=False, data=True, no_progress_bars=False):
    if not data:
        return True
    log = logging.getLogger(__name__.split(".")[-1])
    log.info("Downloading supplementary likelihood data and covmats...")
    return download_github_release(os.path.join(path, "data"), supp_data_name,
                                   supp_data_version, no_progress_bars=no_progress_bars)



# Python 2/3 compatibility
from __future__ import absolute_import, division

# Global
import os
import sys
import signal
from collections import OrderedDict as odict
from pprint import pformat

# Local
from cobaya.yaml import yaml_dump
from cobaya.cosmo_input import input_database
from cobaya.cosmo_input.create_input import create_input
from cobaya.citation import prettyprint_citation, citation
from cobaya.tools import warn_deprecation

try:
    from PySide.QtGui import QWidget, QApplication, QVBoxLayout, QHBoxLayout, QGroupBox
    from PySide.QtGui import QScrollArea, QTabWidget, QComboBox, QPushButton, QTextEdit
    from PySide.QtGui import QFileDialog, QCheckBox, QLabel
    from PySide.QtCore import Slot
except ImportError:
    try:
        from PySide2.QtWidgets import QWidget, QApplication, QVBoxLayout, QHBoxLayout, QGroupBox
        from PySide2.QtWidgets import QScrollArea, QTabWidget, QComboBox, QPushButton, QTextEdit
        from PySide2.QtWidgets import QFileDialog, QCheckBox, QLabel
        from PySide2.QtCore import Slot
    except ImportError:
        QWidget, Slot = object, (lambda: lambda *x: None)

# Quit with C-c
signal.signal(signal.SIGINT, signal.SIG_DFL)


def text(key, contents):
    desc = (contents or {}).get(input_database._desc)
    return desc or key


class MainWindow(QWidget):

    def __init__(self):
        super(MainWindow, self).__init__()
        self.setWindowTitle("Cobaya input generator for Cosmology")
        self.setGeometry(0, 0, 1500, 1000)
        self.move(
            QApplication.desktop().screenGeometry().center() - self.rect().center())
        self.show()
        # Main layout
        self.layout = QHBoxLayout()
        self.setLayout(self.layout)
        self.layout_left = QVBoxLayout()
        self.layout.addLayout(self.layout_left)
        self.layout_output = QVBoxLayout()
        self.layout.addLayout(self.layout_output)
        # LEFT: Options
        self.options = QWidget()
        self.layout_options = QVBoxLayout()
        self.options.setLayout(self.layout_options)
        self.options_scroll = QScrollArea()
        self.options_scroll.setWidget(self.options)
        self.options_scroll.setWidgetResizable(True)
        self.layout_left.addWidget(self.options_scroll)
        titles = odict([
            ["Presets", odict([["preset", "Presets"]])],
            ["Cosmological Model", odict([
                ["theory", "Theory code"],
                ["primordial", "Primordial perturbations"],
                ["geometry", "Geometry"],
                ["hubble", "Hubble parameter constraint"],
                ["matter", "Matter sector"],
                ["neutrinos", "Neutrinos and other extra matter"],
                ["dark_energy", "Lambda / Dark energy"],
                ["bbn", "BBN"],
                ["reionization", "Reionization history"]])],
            ["Data sets", odict([
                ["like_cmb", "CMB experiments"],
                ["like_bao", "BAO experiments"],
                ["like_sn", "SN experiments"],
                ["like_H0", "Local H0 measurements"]])],
            ["Sampler", odict([["sampler", "Samplers"]])]])
        self.combos = odict()
        for group, fields in titles.items():
            group_box = QGroupBox(group)
            self.layout_options.addWidget(group_box)
            group_layout = QVBoxLayout(group_box)
            for a, desc in fields.items():
                self.combos[a] = QComboBox()
                if len(fields) > 1:
                    label = QLabel(desc)
                    group_layout.addWidget(label)
                group_layout.addWidget(self.combos[a])
                self.combos[a].addItems(
                    [text(k, v) for k, v in getattr(input_database, a).items()])
        # PLANCK NAMES CHECKBOX TEMPORARILY DISABLED
        #                if a == "theory":
        #                    # Add Planck-naming checkbox
        #                    self.planck_names = QCheckBox(
        #                        "Keep common parameter names "
        #                        "(useful for fast CLASS/CAMB switching)")
        #                    group_layout.addWidget(self.planck_names)
        # Connect to refreshers -- needs to be after adding all elements
        for field, combo in self.combos.items():
            if field == "preset":
                combo.currentIndexChanged.connect(self.refresh_preset)
            else:
                combo.currentIndexChanged.connect(self.refresh)
        #        self.planck_names.stateChanged.connect(self.refresh_keep_preset)
        # RIGHT: Output + buttons
        self.display_tabs = QTabWidget()
        self.display = {}
        for k in ["yaml", "python", "citations"]:
            self.display[k] = QTextEdit()
            self.display[k].setLineWrapMode(QTextEdit.NoWrap)
            self.display[k].setFontFamily("mono")
            self.display[k].setCursorWidth(0)
            self.display[k].setReadOnly(True)
            self.display_tabs.addTab(self.display[k], k)
        self.layout_output.addWidget(self.display_tabs)
        # Buttons
        self.buttons = QHBoxLayout()
        self.save_button = QPushButton('Save', self)
        self.copy_button = QPushButton('Copy to clipboard', self)
        self.buttons.addWidget(self.save_button)
        self.buttons.addWidget(self.copy_button)
        self.save_button.released.connect(self.save_file)
        self.copy_button.released.connect(self.copy_clipb)
        self.layout_output.addLayout(self.buttons)
        self.save_dialog = QFileDialog()
        self.save_dialog.setFileMode(QFileDialog.AnyFile)
        self.save_dialog.setAcceptMode(QFileDialog.AcceptSave)

    def create_input(self):
        return create_input(
            get_comments=True,
            #           planck_names=self.planck_names.isChecked(),
            **{field: list(getattr(input_database, field).keys())[combo.currentIndex()]
               for field, combo in self.combos.items() if field is not "preset"})

    @Slot()
    def refresh_keep_preset(self):
        self.refresh_display(self.create_input())

    @Slot()
    def refresh(self):
        self.combos["preset"].blockSignals(True)
        self.combos["preset"].setCurrentIndex(0)
        self.combos["preset"].blockSignals(False)
        self.refresh_display(self.create_input())

    @Slot()
    def refresh_preset(self):
        preset = list(getattr(input_database, "preset").keys())[
            self.combos["preset"].currentIndex()]
        info = create_input(
            get_comments=True,
            #            planck_names=self.planck_names.isChecked(),
            preset=preset)
        self.refresh_display(info)
        # Update combo boxes to reflect the preset values, without triggering update
        for k, v in input_database.preset[preset].items():
            if k in [input_database._desc]:
                continue
            self.combos[k].blockSignals(True)
            self.combos[k].setCurrentIndex(
                self.combos[k].findText(
                    text(v, getattr(input_database, k).get(v))))
            self.combos[k].blockSignals(False)

    def refresh_display(self, info):
        try:
            comments = info.pop(input_database._comment, None)
            comments_text = "\n# " + "\n# ".join(comments)
        except (TypeError,  # No comments
                AttributeError):  # Failed to generate info (returned str instead)
            comments_text = ""
        self.display["python"].setText(
            "from collections import OrderedDict\n\ninfo = " +
            pformat(info) + comments_text)
        self.display["yaml"].setText(yaml_dump(info) + comments_text)
        self.display["citations"].setText(prettyprint_citation(citation(info)))

    @Slot()
    def save_file(self):
        ftype = next(k for k, w in self.display.items()
                     if w is self.display_tabs.currentWidget())
        ffilter = {"yaml": "Yaml files (*.yaml *.yml)", "python": "(*.py)",
                   "citations": "(*.txt)"}[ftype]
        fsuffix = {"yaml": ".yaml", "python": ".py", "citations": ".txt"}[ftype]
        fname, path = self.save_dialog.getSaveFileName(
            self.save_dialog, "Save input file", fsuffix, ffilter, os.getcwd())
        if not fname.endswith(fsuffix):
            fname += fsuffix
        with open(fname, "w+") as f:
            f.write(self.display_tabs.currentWidget().toPlainText())

    @Slot()
    def copy_clipb(self):
        self.clipboard.setText(self.display_tabs.currentWidget().toPlainText())


def gui_script():
    warn_deprecation()
    try:
        app = QApplication(sys.argv)
    except NameError:
        # TODO: fix this long logger setup
        from cobaya.log import logger_setup, HandledException
        logger_setup(0, None)
        import logging
        logging.getLogger("cosmo_generator").error(
            "PySide or PySide2 is not installed! "
            "Check Cobaya's documentation for the cosmo_generator "
            "('Basic cosmology runs').")
        raise HandledException
    clip = app.clipboard()
    window = MainWindow()
    window.clipboard = clip
    sys.exit(app.exec_())


if __name__ == '__main__':
    gui_script()



"""
.. module:: tools

:Synopsis: General tools
:Author: Jesus Torrado

"""

# Python 2/3 compatibility
from __future__ import absolute_import, division

# Global
import os
import sys
from copy import deepcopy
from importlib import import_module
import six
import numpy as np  # don't delete: necessary for get_external_function
import scipy.stats as stats  # don't delete: necessary for get_external_function
from collections import OrderedDict as odict
from ast import parse

if six.PY3:
    from inspect import cleandoc, getfullargspec as getargspec
    from math import gcd
    from collections.abc import Mapping
else:
    from inspect import cleandoc, getargspec
    from fractions import gcd
    from collections import Mapping

# Local
from cobaya import __obsolete__
from cobaya.conventions import _package, subfolders, _p_dist, _likelihood, _p_value
from cobaya.log import HandledException

# Logger
import logging

log = logging.getLogger(__name__.split(".")[-1])


# Deepcopy workaround:
def deepcopyfix(olddict):
    if not hasattr(olddict, "keys"):
        return deepcopy(olddict)
    newdict = {}
    for key in olddict:
        if (key == 'theory' or key == 'instance' or key == 'external'):
            newdict[key] = olddict[key]
        else:
            # print(key)
            newdict[key] = deepcopy(olddict[key])
    return newdict


def get_folder(name, kind, sep=os.sep, absolute="True"):
    """
    Gets folder, relative to the package source, of a likelihood, theory or sampler.
    """
    pre = (os.path.dirname(__file__) + sep if absolute
           else "" + (sep if sep == "." else ""))
    return pre + subfolders[kind] + sep + name


def get_class(name, kind=_likelihood):
    """
    Retrieves the requested likelihood (default) or theory class.

    ``info`` must be a dictionary of the kind ``{[class_name]: [options]}``.

    Raises ``ImportError`` if class not found in the appropriate place in the source tree.
    """
    class_folder = get_folder(name, kind, sep=".", absolute=False)
    try:
        return getattr(import_module(class_folder, package=_package), name)
    except:
        if sys.exc_info()[0] == ImportError and str(sys.exc_info()[1]).endswith(name):
            log.error("%s '%s' not found (wrong capitalization?)",
                      kind.capitalize(), name)
            raise HandledException
        else:
            log.error("There was a problem when importing %s '%s':", kind, name)
            raise sys.exc_info()[1]


def get_external_function(string_or_function, name=None):
    """
    Processes an external prior or likelihood, given as a string or a function.

    If the input is a string, it must be evaluable to a function. It can contain import
    statements using :module:`importlib`'s ``import_module``, e.g.
    ``import_module("my_file").my_function``. You can access :module:`scipy.stats` and
    :module:`numpy` members under the handles ``stats`` and ``np`` respectively.

    It is aware of the "value" field for parameters.

    Returns the function.
    """
    if hasattr(string_or_function, "keys"):
        string_or_function = string_or_function.get(_p_value, None)
    if isinstance(string_or_function, six.string_types):
        try:
            if "import_module" in string_or_function:
                sys.path.append(os.path.realpath(os.curdir))
            function = eval(string_or_function)
        except Exception as e:
            log.error("Failed to load external function%s: '%r'",
                      " '%s'" % name if name else "", e)
            raise HandledException
    else:
        function = string_or_function
    if not callable(function):
        log.error("The external function provided " +
                  ("for '%s' " % name if name else "") +
                  "is not an actual function. Got: '%r'", function)
        raise HandledException
    return function


def recursive_odict_to_dict(dictio):
    """
    Recursively converts every ``OrderedDict`` inside the argument into ``dict``.
    """
    if hasattr(dictio, "keys"):
        return {k: recursive_odict_to_dict(v) for k, v in dictio.items()}
    else:
        return dictio


def recursive_update(base, update):
    """
    Recursive dictionary update, from `this stackoverflow question
    <https://stackoverflow.com/questions/3232943>`_.
    Modified for yaml input, where None and {} are almost equivalent
    """
    for update_key, update_value in (update or {}).items():
        update_value = update_value or odict()
        base = base or odict()
        if isinstance(update_value, Mapping):
            base[update_key] = recursive_update(
                base.get(update_key, odict()), update_value)
        else:
            base[update_key] = update_value
    # Trim terminal (o)dicts
    for k, v in (base or {}).items():
        if v in [odict(), {}]:
            base[k] = None
    return base


def make_header(kind, module=None):
    """Creates a header for a particular module of a particular kind."""
    return ("=" * 80).join(["", "\n %s" % kind.title() +
                            (": %s" % module if module else "") + "\n", "\n"])


def ensure_latex(string):
    """Inserts $'s at the beginning and end of the string, if necessary."""
    if string.strip()[0] != r"$":
        string = r"$" + string
    if string.strip()[-1] != r"$":
        string = string + r"$"
    return string


def ensure_nolatex(string):
    """Removes $'s at the beginning and end of the string, if necessary."""
    return string.strip().lstrip("$").rstrip("$")


def read_dnumber(n, d, dtype=float):
    """Reads number as multiples of a given dimension."""
    try:
        if isinstance(n, six.string_types):
            if n[-1].lower() == "d":
                if n.lower() == "d":
                    return d
                return dtype(n[:-1]) * d
            raise ValueError
    except ValueError:
        log.error("Could not convert '%r' to a number.", n)
        raise HandledException
    return n


def prepare_comment(comment):
    """Prepares a string (maybe containing multiple lines) to be written as a comment."""
    return "\n".join(
        ["# " + line.lstrip("#") for line in comment.split("\n") if line]) + "\n"


# Self describing
def is_valid_variable_name(name):
    try:
        parse("%s=None" % name)
        return True
    except SyntaxError:
        return False


def get_scipy_1d_pdf(info):
    """Generates 1d priors from scipy's pdf's from input info."""
    param = list(info.keys())[0]
    info2 = deepcopy(info[param])
    if not info2:
        log.error("No specific prior info given for sampler parameter '%s'." % param)
        raise HandledException
    # What distribution?
    try:
        dist = info2.pop(_p_dist).lower()
    # Not specified: uniform by default
    except KeyError:
        dist = "uniform"
    # Number: uniform with 0 width
    except AttributeError:
        dist = "uniform"
        info2 = {"loc": info2, "scale": 0}
    try:
        pdf_dist = getattr(import_module("scipy.stats", dist), dist)
    except AttributeError:
        log.error("Error creating the prior for parameter '%s': "
                  "The distribution '%s' is unknown to 'scipy.stats'. "
                  "Check the list of allowed possibilities in the docs.", param, dist)
        raise HandledException
    # Recover loc,scale from min,max
    # For coherence with scipy.stats, defaults are min,max=0,1
    if "min" in info2 or "max" in info2:
        if "loc" in info2 or "scale" in info2:
            log.error("You cannot use the 'loc/scale' convention and the 'min/max' "
                      "convention at the same time. Either use one or the other.")
            raise HandledException
        minmaxvalues = {"min": 0, "max": 1}
        for limit in minmaxvalues:
            try:
                value = info2.pop(limit, minmaxvalues[limit])
                minmaxvalues[limit] = np.float(value)
            except (TypeError, ValueError):
                log.error("Invalid value '%s: %r' in param '%s' (it must be a number)",
                          limit, value, param)
                raise HandledException
        info2["loc"] = minmaxvalues["min"]
        info2["scale"] = minmaxvalues["max"] - minmaxvalues["min"]
    # Check for improper priors
    if not np.all(np.isfinite([info2.get(x, 0) for x in ["loc", "scale", "min", "max"]])):
        log.error("Improper prior for parameter '%s'.", param)
        raise HandledException
    # Generate and return the frozen distribution
    try:
        return pdf_dist(**info2)
    except TypeError as tp:
        log.error(
            "'scipy.stats' produced an error: <<%r>>. "
            "This probably means that the distribution '%s' "
            "does not recognize the parameter mentioned in the 'scipy' error above.",
            str(tp), dist)
        raise HandledException


def _fast_uniform_logpdf(self, x):
    """WARNING: logpdf(nan) = -inf"""
    if not hasattr(self, "_cobaya_mlogscale"):
        self._cobaya_mlogscale = -np.log(self.kwds["scale"])
        self._cobaya_max = self.kwds["loc"] + self.kwds["scale"]
    x_ = np.array(x)
    return np.where(np.logical_and(x_ >= self.kwds["loc"], x_ <= self._cobaya_max),
                    self._cobaya_mlogscale, -np.inf)


def _fast_norm_logpdf(self, x):
    """WARNING: logpdf(nan) = -inf"""
    if not hasattr(self, "_cobaya_mlogscale"):
        self._cobaya_mlogscale = -np.log(self.kwds["scale"])
    x_ = (np.array(x) - self.kwds["loc"]) / self.kwds["scale"]
    return self.dist._logpdf(x_) + self._cobaya_mlogscale


def KL_norm(m1=None, S1=np.array([]), m2=None, S2=np.array([])):
    """Kullback-Leibler divergence between 2 gaussians."""
    S1, S2 = [np.atleast_2d(S) for S in [S1, S2]]
    assert S1.shape[0], "Must give at least S1"
    dim = S1.shape[0]
    if m1 is None:
        m1 = np.zeros(dim)
    if not S2.shape[0]:
        S2 = np.identity(dim)
    if m2 is None:
        m2 = np.zeros(dim)
    S2inv = np.linalg.inv(S2)
    KL = 0.5 * (np.trace(S2inv.dot(S1)) + (m1 - m2).dot(S2inv).dot(m1 - m2) -
                dim + np.log(np.linalg.det(S2) / np.linalg.det(S1)))
    return KL


def compare_params_lists(list_A, list_B):
    """
    Compares two parameter lists, and returns a dict with the following keys
    (only present if applicable):

      `duplicates_[A|B]`: duplicate elemnts in list 1|2
    """
    result = {}
    names = ["A", "B"]
    # Duplicates
    list_A_copy, list_B_copy = list_A[:], list_B[:]
    for n, l in zip(names, [list_A_copy, list_B_copy]):
        [l.pop(i) for i in sorted([l.index(x) for x in set(l)])[::-1]]
        if l:
            result["duplicate_%s" % n] = list(set(l))
    sets = {"A": set(list_A), "B": set(list_B)}
    for n1, n2 in [["A", "B"], ["B", "A"]]:
        missing = sets[n1].difference(sets[n2])
        if missing:
            result["%s_but_not_%s" % (n1, n2)] = list(missing)
    return result


def relative_to_int(numbers, precision=1 / 10):
    """
    Turns relative numbers (e.g. relative speeds) into integer,
    up to some given `precision` on differences.
    """
    numbers = np.array(np.round(np.array(numbers) / min(numbers) / precision), dtype=int)
    return np.array(
        numbers / np.ufunc.reduce(np.frompyfunc(gcd, 2, 1), numbers), dtype=int)


def create_banner(msg):
    """
    Puts message into an attention-grabbing banner.
    """
    msg_clean = cleandoc(msg)
    longest_line_len = max([len(line) for line in msg_clean.split("\n")])
    return ("*" * longest_line_len + "\n" + msg_clean + "\n" + "*" * longest_line_len)


def warn_deprecation_python2():
    msg = """
    *WARNING*: Python 2 support will eventually be dropped
    (it is already unsupported by many scientific Python modules).

    Please use Python 3!

    In some systems, the Python 3 command may be python3 instead of python.
    If that is the case, use pip3 instead of pip to install Cobaya.
    """
    if not six.PY3:
        print(create_banner(msg))


def warn_deprecation_version():
    msg = """
    *WARNING*: You are using an obsolete version of Cobaya, which is no loger maintained.

    Please, update asap to the last version (e.g. with ``pip install cobaya --upgrade``).
    """
    if __obsolete__:
        print(create_banner(msg))


def warn_deprecation():
    warn_deprecation_python2()
    warn_deprecation_version()


def progress_bar(logger, percentage, final_text=""):
    """Very simple, multiline, logger-compatible progress bar, with increments of 5%."""
    progress = int(percentage/5)
    logger.info(" |%s| %3d%% %s",
                "@" * progress + "-" * (20-progress), percentage, final_text)



"""
.. module:: yaml

:Synopsis: Custom YAML loader and dumper
:Author: Jesus Torrado (parts of the code comes from stackoverflow user's)

Customization of YAML's loaded and dumper:

1. Matches 1e2 as 100 (no need for dot, or sign after e),
   from https://stackoverflow.com/a/30462009
2. Wrapper to load mappings as OrderedDict (for likelihoods and params),
   from https://stackoverflow.com/a/21912744

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
import yaml
import re
from collections import OrderedDict as odict
import numpy as np

# Local
from cobaya.tools import prepare_comment


# Exceptions

class InputSyntaxError(Exception):
    """Syntax error in YAML input."""


class OutputError(Exception):
    """Error when dumping YAML info."""


def yaml_load(text_stream, Loader=yaml.Loader, object_pairs_hook=odict, file_name=None):
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping)
    OrderedLoader.add_implicit_resolver(
        u'tag:yaml.org,2002:float',
        re.compile(u'''^(?:
            [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
            |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
            |\\.[0-9_]+(?:[eE][-+][0-9]+)?
            |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
            |[-+]?\\.(?:inf|Inf|INF)
            |\\.(?:nan|NaN|NAN))$''', re.X),
        list(u'-+0123456789.'))
    try:
        return yaml.load(text_stream, OrderedLoader)
    # Redefining the general exception to give more user-friendly information
    except (yaml.YAMLError, TypeError) as exception:
        errstr = "Error in your input file " + ("'" + file_name + "'" if file_name else "")
        if hasattr(exception, "problem_mark"):
            line = 1 + exception.problem_mark.line
            column = 1 + exception.problem_mark.column
            signal = " --> "
            signal_right = "    <---- "
            sep = "|"
            context = 4
            lines = text_stream.split("\n")
            pre = ((("\n" + " " * len(signal) + sep).join(
                [""] + lines[max(line - 1 - context, 0):line - 1]))) + "\n"
            errorline = (signal + sep + lines[line - 1] +
                         signal_right + "column %s" % column)
            post = ((("\n" + " " * len(signal) + sep).join(
                [""] + lines[line + 1 - 1:min(line + 1 + context - 1, len(lines))]))) + "\n"
            raise InputSyntaxError(
                errstr + " at line %d, column %d." % (line, column) +
                pre + errorline + post +
                "Maybe inconsistent indentation, '=' instead of ':', "
                "no space after ':', or a missing ':' on an empty group?")
        else:
            raise InputSyntaxError(errstr)


def yaml_load_file(file_name):
    """Wrapper to load a yaml file."""
    with open(file_name, "r") as file:
        lines = "".join(file.readlines())
    return yaml_load(lines, file_name=file_name)


def yaml_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
    class OrderedDumper(Dumper):
        pass

    # Dump OrderedDict's as plain dictionaries, but keeping the order
    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items())

    OrderedDumper.add_representer(odict, _dict_representer)

    # Dump tuples as yaml "sequences"
    def _tuple_representer(dumper, data):
        return dumper.represent_sequence(
            yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG, list(data))

    OrderedDumper.add_representer(tuple, _tuple_representer)

    # Numpy arrays and numbers
    def _numpy_array_representer(dumper, data):
        return dumper.represent_sequence(
            yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG, data.tolist())

    OrderedDumper.add_representer(np.ndarray, _numpy_array_representer)

    def _numpy_int_representer(dumper, data):
        return dumper.represent_int(data)

    OrderedDumper.add_representer(np.int64, _numpy_int_representer)

    def _numpy_float_representer(dumper, data):
        return dumper.represent_float(data)

    OrderedDumper.add_representer(np.float64, _numpy_float_representer)

    # Dummy representer that prints True for non-representable python objects
    # (prints True instead of nothing because some functions try cast values to bool)
    def _null_representer(dumper, data):
        return dumper.represent_scalar('tag:yaml.org,2002:bool', 'true')

    OrderedDumper.add_representer(type(lambda: None), _null_representer)
    OrderedDumper.add_multi_representer(object, _null_representer)

    # Dump!
    return yaml.dump(data, stream, OrderedDumper, **kwds)


def yaml_dump_file(file_name, data, comment=None, error_if_exists=True):
    if error_if_exists and os.path.isfile(file_name):
        raise IOError("File exists: '%s'" % file_name)
    with open(file_name, "w+") as f:
        if comment:
            f.write(prepare_comment(comment))
        f.write(yaml_dump(data))
